# @license
# Copyright 2017 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import credentials_provider, google_credentials

default_credentials_manager = credentials_provider.CredentialsManager()
default_credentials_manager.register(
    u'google-brainmaps',
    lambda _parameters: google_credentials.GoogleCredentialsProvider(
        client_id=u'639403125587-ue3c18dalqidqehs1n1p5rjvgni5f7qu.apps.googleusercontent.com',
        client_secret=u'kuaqECaVXOKEJ2L6ifZu4Aqt',
        scopes=[u'https://www.googleapis.com/auth/brainmaps'],
    ))



# @license
# Copyright 2016 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import, division, print_function

import collections
import threading

import numpy as np

from . import downsample, downsample_scales
from .chunks import encode_jpeg, encode_npz, encode_raw
from . import trackable_state
from .random_token import make_random_token


class MeshImplementationNotAvailable(Exception):
    pass

class MeshesNotSupportedForVolume(Exception):
    pass

class InvalidObjectIdForMesh(Exception):
    pass

DownsamplingScaleInfo = collections.namedtuple('DownsamplingScaleInfo', ['key',
                                                                         'downsample_factor',
                                                                         'voxel_size',
                                                                         'shape', ])


def get_scale_key(scale):
    return '%d,%d,%d' % scale


class LocalVolume(trackable_state.ChangeNotifier):
    def __init__(self,
                 data,
                 offset=None,
                 voxel_offset=None,
                 skeletons=None,
                 voxel_size=(1, 1, 1),
                 encoding='npz',
                 max_voxels_per_chunk_log2=None,
                 volume_type=None,
                 mesh_options=None,
                 downsampling='3d',
                 max_downsampling=downsample_scales.DEFAULT_MAX_DOWNSAMPLING,
                 max_downsampled_size=downsample_scales.DEFAULT_MAX_DOWNSAMPLED_SIZE,
                 max_downsampling_scales=downsample_scales.DEFAULT_MAX_DOWNSAMPLING_SCALES):
        """Initializes a LocalVolume.

        @param data: 3-d [z, y, x] array or 4-d [channel, z, y, x] array.

        @param downsampling: '3d' to use isotropic downsampling, '2d' to downsample separately in
            XY, XZ, and YZ, None to use no downsampling.

        @param max_downsampling: Maximum amount by which on-the-fly downsampling may reduce the
            volume of a chunk.  For example, 4x4x4 downsampling reduces the volume by 64.

        @param volume_type: either 'image' or 'segmentation'.  If not specified, guessed from the
            data type.

        @param voxel_size: Sequence [x, y, z] of floats.  Specifies the voxel size.

        @param mesh_options: A dict with the following keys specifying options for mesh
            simplification for 'segmentation' volumes:

                - max_quadrics_error: float.  Edge collapses with a larger associated quadrics error
                  than this amount are prohibited.  Set this to a negative number to disable mesh
                  simplification, and just use the original mesh produced by the marching cubes
                  algorithm.  Defaults to 1e6.  The effect of this value depends on the voxel_size.

                - max_normal_angle_deviation: float.  Edge collapses that change a triangle normal
                  by more than this angle are prohibited.  The angle is specified in degrees.
                  Defaults to 90.

                - lock_boundary_vertices: bool.  Retain all vertices along mesh surface boundaries,
                  which can only occur at the boundary of the volume.  Defaults to true.
        """
        super(LocalVolume, self).__init__()
        if hasattr(data, 'attrs'):
            if 'resolution' in data.attrs:
                if voxel_size is None:
                    voxel_size = tuple(data.attrs['resolution'])[::-1]
                if 'offset' in data.attrs:
                    if offset is None and voxel_offset is None:
                        offset = tuple(data.attrs['offset'])[::-1]
        voxel_size = np.array(voxel_size)
        self.token = make_random_token()
        self.max_voxels_per_chunk_log2 = max_voxels_per_chunk_log2
        self.data = data
        self.skeletons = skeletons
        if voxel_offset is not None:
            if offset is not None:
                raise ValueError('Must specify at most one of \'offset\' and \'voxel_offset\'.')
            voxel_offset = np.array(voxel_offset)
            offset = tuple(voxel_offset * voxel_size)
        if offset is None:
            offset = (0, 0, 0)
        self.offset = tuple(offset)
        self.data_type = np.dtype(data.dtype).name
        if self.data_type == 'float64':
            self.data_type = 'float32'
        self.encoding = encoding
        if len(data.shape) == 3:
            self.num_channels = 1
            original_shape = data.shape[::-1]
        else:
            if len(data.shape) != 4:
                raise ValueError('data array must be 3- or 4-dimensional.')
            self.num_channels = data.shape[0]
            original_shape = data.shape[1:][::-1]
        original_shape = np.array(original_shape)
        if volume_type is None:
            if self.num_channels == 1 and (self.data_type == 'uint16' or
                                           self.data_type == 'uint32' or
                                           self.data_type == 'uint64'):
                volume_type = 'segmentation'
            else:
                volume_type = 'image'
        self.volume_type = volume_type

        self._mesh_generator = None
        self._mesh_generator_pending = None
        self._mesh_generator_lock = threading.Condition()
        self._mesh_options = mesh_options.copy() if mesh_options is not None else dict()


        voxel_size = np.array(voxel_size)
        self.voxel_size = voxel_size

        self.two_dimensional_scales = None
        self.three_dimensional_scales = None

        if downsampling is None:
            downsampling_scales = self.three_dimensional_scales = [(1, 1, 1)]
        elif downsampling == '3d':
            self.three_dimensional_scales = downsample_scales.compute_near_isotropic_downsampling_scales(
                size=original_shape,
                voxel_size=voxel_size,
                dimensions_to_downsample=[0, 1, 2],
                max_downsampling=max_downsampling,
                max_downsampled_size=max_downsampled_size,
                max_scales=max_downsampling_scales)
            downsampling_scales = self.three_dimensional_scales
        elif downsampling == '2d':
            self.two_dimensional_scales = downsample_scales.compute_two_dimensional_near_isotropic_downsampling_scales(
                size=original_shape,
                voxel_size=voxel_size,
                max_downsampling=max_downsampling,
                max_downsampled_size=max_downsampled_size,
                max_scales=max_downsampling_scales)
            downsampling_scales = [s for level in self.two_dimensional_scales for s in level]
        downsampling_scale_info = self.downsampling_scale_info = {}
        for scale in downsampling_scales:
            info = DownsamplingScaleInfo(key=get_scale_key(scale),
                                         voxel_size=tuple(voxel_size * scale),
                                         downsample_factor=scale,
                                         shape=tuple(np.cast[int](np.ceil(original_shape / scale))))
            downsampling_scale_info[info.key] = info

    def info(self):
        info = dict(volumeType=self.volume_type,
                    dataType=self.data_type,
                    encoding=self.encoding,
                    numChannels=self.num_channels,
                    generation=self.change_count,
        )
        if self.max_voxels_per_chunk_log2 is not None:
            info['maxVoxelsPerChunkLog2'] = self.max_voxels_per_chunk_log2

        def get_scale_info(s):
            info = self.downsampling_scale_info[get_scale_key(s)]
            return dict(key=info.key,
                        offset=self.offset,
                        sizeInVoxels=info.shape,
                        voxelSize=info.voxel_size)

        if self.two_dimensional_scales is not None:
            info['twoDimensionalScales'] = [[get_scale_info(s) for s in level]
                                            for level in self.two_dimensional_scales]
        if self.three_dimensional_scales is not None:
            info['threeDimensionalScales'] = [get_scale_info(s)
                                              for s in self.three_dimensional_scales]
        if self.skeletons is not None:
            info['skeletonVertexAttributes'] = self.skeletons.get_vertex_attributes_spec()
        return info

    def get_encoded_subvolume(self, data_format, start, end, scale_key='1,1,1'):
        scale_info = self.downsampling_scale_info.get(scale_key)
        if scale_info is None:
            raise ValueError('Invalid scale.')
        shape = scale_info.shape
        downsample_factor = scale_info.downsample_factor
        for i in range(3):
            if end[i] < start[i] or start[i] < 0 or end[i] > shape[i]:
                raise ValueError('Out of bounds data request.')

        indexing_expr = tuple(np.s_[start[i] * downsample_factor[i]:end[i] * downsample_factor[i]]
                              for i in (2, 1, 0))
        if len(self.data.shape) == 3:
            full_downsample_factor = downsample_factor[::-1]
            subvol = self.data[indexing_expr]
        else:
            full_downsample_factor = (1, ) + downsample_factor[::-1]
            subvol = self.data[(np.s_[:], ) + indexing_expr]
        if subvol.dtype == 'float64':
            subvol = np.cast[np.float32](subvol)

        if downsample_factor != (1, 1, 1):
            if self.volume_type == 'image':
                subvol = downsample.downsample_with_averaging(subvol, full_downsample_factor)
            else:
                subvol = downsample.downsample_with_striding(subvol, full_downsample_factor)
        content_type = 'application/octet-stream'
        if data_format == 'jpeg':
            data = encode_jpeg(subvol)
            content_type = 'image/jpeg'
        elif data_format == 'npz':
            data = encode_npz(subvol)
        elif data_format == 'raw':
            data = encode_raw(subvol)
        else:
            raise ValueError('Invalid data format requested.')
        return data, content_type

    def get_object_mesh(self, object_id):
        mesh_generator = self._get_mesh_generator()
        data = mesh_generator.get_mesh(object_id)
        if data is None:
            raise InvalidObjectIdForMesh()
        return data

    def _get_mesh_generator(self):
        if self._mesh_generator is not None:
            return self._mesh_generator
        while True:
            with self._mesh_generator_lock:
                if self._mesh_generator is not None:
                    return self._mesh_generator
                if self._mesh_generator_pending is not None:
                    while self._mesh_generator is None:
                        self._mesh_generator_lock.wait()
                    if self._mesh_generator is not None:
                        return self._mesh_generator
                try:
                    from . import _neuroglancer
                except ImportError:
                    raise MeshImplementationNotAvailable()
                if not (self.num_channels == 1 and
                        (self.data_type == 'uint8' or self.data_type == 'uint16' or
                         self.data_type == 'uint32' or self.data_type == 'uint64')):
                    raise MeshesNotSupportedForVolume()
                pending_obj = object()
                self._mesh_generator_pending = pending_obj
            if len(self.data.shape) == 4:
                data = self.data[0, :, :, :]
            else:
                data = self.data
            new_mesh_generator = _neuroglancer.OnDemandObjectMeshGenerator(
                data, self.voxel_size, self.offset / self.voxel_size, **self._mesh_options)
            with self._mesh_generator_lock:
                if self._mesh_generator_pending is not pending_obj:
                    continue
                self._mesh_generator = new_mesh_generator
                self._mesh_generator_pending = False
                self._mesh_generator_lock.notify_all()
            return new_mesh_generator

    def __deepcopy__(self, memo):
        """Since this type is immutable, we don't need to deepcopy it.

        Actually deep copying would intefere with the use of deepcopy by JsonObjectWrapper.
        """
        return self

    def invalidate(self):
        """Mark the data invalidated.  Clients will refetch the volume."""
        with self._mesh_generator_lock:
            self._mesh_generator_pending = None
            self._mesh_generator = None
        self._dispatch_changed_callbacks()



# @license
# Copyright 2017 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import base64
import collections
import numbers
import traceback

import numpy as np
import six

from . import viewer_state
from .json_wrappers import (JsonObjectWrapper, array_wrapper, optional, text_type, typed_list,
                            typed_set, typed_string_map, wrapped_property)

_uint64_keys = frozenset(['t', 'v'])
_map_entry_keys = frozenset(['key', 'value'])

MapEntry = collections.namedtuple('MapEntry', ['key', 'value'])

def layer_selected_value(x):
    if isinstance(x, numbers.Number):
        return x
    if isinstance(x, dict):
        if six.viewkeys(x) == _uint64_keys and x.get('t') == 'u64':
            return int(x['v'])
        if six.viewkeys(x) == _map_entry_keys:
            return MapEntry(int(x['key']), int(x['value']))

LayerSelectedValues = typed_string_map(layer_selected_value)


class ScreenshotReply(JsonObjectWrapper):
    __slots__ = ()
    id = wrapped_property('id', text_type)
    image = wrapped_property('image', base64.b64decode)
    image_type = imageType = wrapped_property('imageType', text_type)

class ActionState(JsonObjectWrapper):
    __slots__ = ()
    viewer_state = viewerState = wrapped_property('viewerState', viewer_state.ViewerState)
    selected_values = selectedValues = wrapped_property('selectedValues', LayerSelectedValues)
    mouse_spatial_coordinates = mouseSpatialCoordinates = wrapped_property(
        'mouseSpatialCoordinates', optional(array_wrapper(np.float32, 3)))
    mouse_voxel_coordinates = mouseVoxelCoordinates = wrapped_property(
        'mouseVoxelCoordinates', optional(array_wrapper(np.float32, 3)))
    screenshot = wrapped_property('screenshot', optional(ScreenshotReply))


class Actions(object):
    def __init__(self, set_config):
        self._action_handlers = dict()
        self._set_config = set_config

    def add(self, name, handler):
        self._action_handlers.setdefault(name, set()).add(handler)
        self._update_config()

    def clear(self):
        self._action_handlers.clear()
        self._update_config()

    def remove(self, name, handler):
        handlers = self._action_handlers.get(name)
        if handlers is None:
            return
        handlers.remove(handler)
        if not handlers:
            del self._action_handlers[name]
        self._update_config()

    def _update_config(self):
        self._set_config(six.viewkeys(self._action_handlers))

    def invoke(self, name, state):
        state = ActionState(state)
        handlers = self._action_handlers.get(name)
        if handlers is not None:
            for handler in handlers:
                try:
                    handler(state)
                except:
                    traceback.print_exc()

EventActionMap = typed_string_map(text_type)

class InputEventBindings(JsonObjectWrapper):
    __slots__ = ()
    viewer = wrapped_property('viewer', EventActionMap)
    slice_view = sliceView = wrapped_property('sliceView', EventActionMap)
    perspective_view = perspectiveView = wrapped_property('perspectiveView', EventActionMap)
    data_view = dataView = wrapped_property('dataView', EventActionMap)


class PrefetchState(JsonObjectWrapper):
    __slots__ = ()
    supports_validation = True
    priority = wrapped_property('priority', optional(int, 0))
    state = wrapped_property('state', viewer_state.ViewerState)


class ConfigState(JsonObjectWrapper):
    __slots__ = ()
    credentials = wrapped_property('credentials', typed_string_map(dict))
    actions = wrapped_property('actions', typed_set(text_type))
    input_event_bindings = inputEventBindings = wrapped_property('inputEventBindings',
                                                                 InputEventBindings)
    status_messages = statusMessages = wrapped_property('statusMessages',
                                                        typed_string_map(text_type))
    source_generations = sourceGenerations = wrapped_property('sourceGenerations',
                                                              typed_string_map(int))
    screenshot = wrapped_property('screenshot', optional(text_type))
    show_ui_controls = showUIControls = wrapped_property('showUIControls', optional(bool, True))
    show_location = showLocation = wrapped_property('showLocation', optional(bool, True))
    show_layer_panel = showLayerPanel = wrapped_property('showLayerPanel', optional(bool, True))
    show_help_button = showHelpButton = wrapped_property('showHelpButton', optional(bool, True))
    show_panel_borders = showPanelBorders = wrapped_property('showPanelBorders',
                                                             optional(bool, True))
    viewer_size = viewerSize = wrapped_property('viewerSize', optional(array_wrapper(np.int64, 2)))
    prefetch = wrapped_property('prefetch', typed_list(PrefetchState))


class PrivateState(JsonObjectWrapper):
    __slots__ = ()
    credentials = wrapped_property('credentials', typed_string_map(optional(int)))



# @license
# Copyright 2018 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for merge_tool.py"""

from __future__ import absolute_import

import unittest

from . import merge_tool


class BlockMaskTest(unittest.TestCase):
    def test_basic(self):

        mask = merge_tool.BlockMask()
        mask.add(0, (5, 3, 1))

        self.assertEqual(mask.blocks, [{
            (5, 3, 1): 1
        }, {
            (2, 1, 0): 1
        }, {
            (1, 0, 0): 1
        }, {
            (0, 0, 0): 1
        }])

        mask.add(0, (5, 3, 0))
        self.assertEqual(mask.blocks, [
            {
                (5, 3, 0): 1,
                (5, 3, 1): 1
            },
            {
                (2, 1, 0): 2
            },
            {
                (1, 0, 0): 2
            },
            {
                (0, 0, 0): 2
            },
        ])

        mask.add(0, (5, 2, 1))
        mask.add(0, (5, 2, 0))

        mask.add(0, (4, 2, 1))
        mask.add(0, (4, 2, 0))
        mask.add(0, (4, 3, 1))
        self.assertEqual(mask.blocks, [
            {
                (4, 2, 1): 1,
                (4, 2, 0): 1,
                (4, 3, 1): 1,
                (5, 2, 0): 1,
                (5, 2, 1): 1,
                (5, 3, 0): 1,
                (5, 3, 1): 1
            },
            {
                (2, 1, 0): 7
            },
            {
                (1, 0, 0): 7
            },
            {
                (0, 0, 0): 7
            },
        ])

        mask.add(0, (4, 3, 0))
        self.assertEqual(mask.blocks, [
            {},
            {
                (2, 1, 0): 8
            },
            {
                (1, 0, 0): 8
            },
            {
                (0, 0, 0): 8
            },
        ])

        mask.remove(0, (4, 3, 0))
        self.assertEqual(mask.blocks, [
            {
                (4, 2, 1): 1,
                (4, 2, 0): 1,
                (4, 3, 1): 1,
                (5, 2, 0): 1,
                (5, 2, 1): 1,
                (5, 3, 0): 1,
                (5, 3, 1): 1
            },
            {
                (2, 1, 0): 7
            },
            {
                (1, 0, 0): 7
            },
            {
                (0, 0, 0): 7
            },
        ])

        mask.remove(1, (2, 1, 0))
        self.assertEqual(mask.blocks, [{}, {}, {}, {}])


if __name__ == '__main__':
    unittest.main()



# @license
# Copyright 2017 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for equivalence_map.py"""

from __future__ import absolute_import

import unittest

from . import equivalence_map


class EquivalenceMapTest(unittest.TestCase):
    def test_basic(self):

        m = equivalence_map.EquivalenceMap()

        for i in range(24):
            self.assertEqual(0, m[i])
            self.assertEqual(i + 1, m[i + 1])
            self.assertEqual(set([i + 1]), set(m.members(i + 1)))
            result = m.union(i, i + 1)
            self.assertEqual(result, 0)
            self.assertEqual(0, m[i])
            self.assertEqual(0, m[i + 1])
            self.assertEqual(set(range(i + 2)), set(m.members(i)))

        for i in range(25, 49):
            self.assertEqual(25, m[i])
            self.assertEqual(i + 1, m[i + 1])
            result = m.union(i, i + 1)
            self.assertEqual(result, 25)
            self.assertEqual(25, m[i])
            self.assertEqual(25, m[i + 1])
            self.assertEqual(set(range(25, i + 2)), set(m.members(i)))
        self.assertNotEqual(m[15], m[40])
        result = m.union(15, 40)
        self.assertEqual(0, result)
        self.assertEqual(0, m[15])
        self.assertEqual(0, m[40])
        self.assertEqual(set(range(50)), set(m.members(15)))

        for i in range(50):
            self.assertEqual(0, m[i])
            self.assertEqual(set(range(50)), set(m.members(i)))

        for i in range(51, 100):
            self.assertEqual(set([i]), set(m.members(i)))

    def test_init_simple(self):
        m = equivalence_map.EquivalenceMap([[1, 2, 3], [4, 5]])
        self.assertEqual(1, m[1])
        self.assertEqual(1, m[2])
        self.assertEqual(1, m[3])
        self.assertEqual(4, m[4])
        self.assertEqual(4, m[5])
        self.assertEqual(set([1, 2, 3, 4, 5]), set(m.keys()))
        self.assertEqual([[1, 2, 3], [4, 5]], m.to_json())

    def test_delete_set(self):
        m = equivalence_map.EquivalenceMap([[1, 2, 3], [4, 5]])
        m.delete_set(5)
        self.assertEqual([[1, 2, 3]], m.to_json())

    def test_isolate_element(self):
        m = equivalence_map.EquivalenceMap([[1, 2, 3], [4, 5]])
        m.isolate_element(1)
        self.assertEqual([[2, 3], [4, 5]], m.to_json())

        m.isolate_element(1)
        self.assertEqual([[2, 3], [4, 5]], m.to_json())


if __name__ == '__main__':
    unittest.main()



# @license
# Copyright 2017 Google Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import unittest

from . import url_state


class ConvertStringLiteralTest(unittest.TestCase):
    def test_basic(self):
        quote_initial=u'\''
        desired_quote_char = u'"'
        quote_search = url_state.DOUBLE_QUOTE_PATTERN

        self.assertEqual(
            url_state._convert_string_literal(
                u"'hello'",
                quote_initial=quote_initial,
                quote_replace=desired_quote_char,
                quote_search=quote_search), u"\"hello\"")

        self.assertEqual(
            url_state._convert_string_literal(
                u"'hello\"foo'",
                quote_initial=quote_initial,
                quote_replace=desired_quote_char,
                quote_search=quote_search), u"\"hello\\\"foo\"")

class UrlSafeToJson(unittest.TestCase):
    def test_basic(self):
        self.assertEqual(url_state.url_safe_to_json("""{'a':'b'_'b':'c'}"""), """{"a":"b","b":"c"}""")
        self.assertEqual(url_state.url_safe_to_json("""['a'_true]"""), """["a",true]""")
        self.assertEqual(url_state.url_safe_to_json("""['a',true]"""), """["a",true]""")


if __name__ == '__main__':
    unittest.main()
